# file      : build/bootstrap.make
# author    : Boris Kolpackov <boris@kolpackov.net>
# copyright : Copyright (c) 2004 Boris Kolpackov
# license   : GNU GPL v2; http://www.gnu.org/licenses/gpl.txt

# Configure make.
#

# Forces make to delete targets whos rebuild commands failed but
# updated the target.
#
.DELETE_ON_ERROR:

# Remove all default rules.
#
.SUFFIXES:

ifeq ($(filter -r,$(MAKEFLAGS)),)
MAKEFLAGS += -r
endif


# Disable implicit phony feature.
#
ifneq ($(findstring --no-implicit-phony,$(MAKEFLAGS)),)
MAKEFLAGS += --no-implicit-phony
endif



ifndef %makefile%
%replica%     :=
# @@ maybe I should set makefile to bootstrap and restore it at the end.
#    This way I will be able to use %makefile% below (instead of _self) and
#    also it will be consistent.
#
%makefile%    := $(realpath $(firstword $(MAKEFILE_LIST)))
%interactive% := true

# include facility
#
#@@ I am doing realpath on the whole MAKEFILE_LIST. What about
#   caching this in some system variable?
#
#@@ Need a more reliable than findstring function to look for a
#   string in a list.
#
#@@ It seems realpath returns empty string if path does not exist.
#   Is it always what I want?
#

define include-body
%interactive% :=
%makefile%    := $(realpath $1)
%replica%     := $(if $(patsubst 0,,$(words $(filter $(%makefile%),$(realpath $(MAKEFILE_LIST))))),true,)

$(if $5,$$(if $$(%makefile%),$$(eval %include_once_$$(%makefile%)% += $5)))

include $1

%makefile%    := $2
%interactive% := $3
%replica%     := $4
endef

define include
$(eval $(call include-body,$1,$(%makefile%),$(%interactive%),$(%replica%)))
endef

define include-once-file-filter
$(eval $(if $(findstring $(realpath $1),$(realpath $(MAKEFILE_LIST))),,\
$(call include-body,$1,$(%makefile%),$(%interactive%),$(%replica%))))
endef


define include-once-value-filter
$(eval $(if $(findstring $2,$(value %include_once_$(realpath $1)%)),,\
$(call include-body,$1,$(%makefile%),$(%interactive%),$(%replica%),$2)))
endef

define include-once
$(eval $(if $2,$(call include-once-value-filter,$1,$(strip $2)),\
$(call include-once-file-filter,$1)))
endef

#
###
#

define -include-body
%interactive% :=
%makefile%    := $(realpath $1)
%replica%     := $(if $(patsubst 0,,$(words $(filter $(%makefile%),$(realpath $(MAKEFILE_LIST))))),true,)

$(if $5,$$(if $$(%makefile%),$$(eval %include_once_$$(%makefile%)% += $5)))

-include $1

%makefile%    := $2
%interactive% := $3
%replica%     := $4
endef

define -include
$(eval $(call -include-body,$1,$(%makefile%),$(%interactive%),$(%replica%)))
endef

define -include-once-file-filter
$(eval $(if $(findstring $(realpath $1),$(realpath $(MAKEFILE_LIST))),,\
$(call -include-body,$1,$(%makefile%),$(%interactive%),$(%replica%))))
endef


define -include-once-value-filter
$(eval $(if $(findstring $2,$(value %include_once_$(realpath $1)%)),,\
$(call -include-body,$1,$(%makefile%),$(%interactive%),$(%replica%),$2)))
endef

define -include-once
$(eval $(if $2,$(call -include-once-value-filter,$1,$(strip $2)),\
$(call -include-once-file-filter,$1)))
endef


#$(warning include facility is up)

endif

bld_root := $(realpath $(dir $(lastword $(MAKEFILE_LIST))))

#$(warning namespace facility is up)

# frame facility
#
$(call include-once,$(bld_root)/frame.make)

# import facility
#
$(call include-once,$(bld_root)/import.make)

# load facility
#
$(call include-once,$(bld_root)/load.make)


# initialize {src,out}_{root,base}
#
define find-bootstrap-base
$(if $(findstring $(lastword $1),$(%makefile%)),$2,\
$(call find-bootstrap-base,\
$(wordlist 1,$(words $(wordlist 2,$(words $1),$1)),$1),\
$(lastword $1)))
endef


src_root := $(realpath $(dir $(call find-bootstrap-base,\
$(realpath $(MAKEFILE_LIST))))..)

src_base := $(realpath $(dir $(%makefile%)))

ifneq ($(src_root),$(src_base))

_rel_path := $(subst $(src_root)/,,$(src_base))

ifeq ($(origin out_root),undefined)
  out_root := $(realpath $(patsubst %$(_rel_path),%,$(CURDIR)))
endif

out_base := $(out_root)/$(_rel_path)

else

ifeq ($(origin out_root),undefined)
  out_root := $(realpath $(CURDIR))
endif

out_base := $(out_root)

endif

scf_root := $(src_root)/build
dcf_root := $(out_root)/build

$(out_root)/%: out_root := $(out_root)
$(out_root)/%: src_root := $(src_root)
$(out_root)/%: scf_root := $(scf_root)
$(out_root)/%: dcf_root := $(dcf_root)
$(out_root)/%: project_name := $(project_name)

$(out_base)/%: out_root := $(out_root)
$(out_base)/%: out_base := $(out_base)
$(out_base)/%: src_base := $(src_base)

ifdef build_debug

$(warning "bld root: $(bld_root)")
$(warning "scf root: $(scf_root)")
$(warning "dcf root: $(dcf_root)")
$(warning "src root: $(src_root)")
$(warning "src base: $(src_base)")
$(warning "out root: $(out_root)")
$(warning "out base: $(out_base)")

endif

#$(warning paths are set)

# disfigure target
#
.PHONY: disfigure

#@@ Can't do it here because of the target hijacking.
#
#disfigure:: clean

$(call include-once,$(bld_root)/message.make)
$(call include-once,$(bld_root)/dir.make)
